# Generated from D:/Projects/BESSER/besser/BUML/notations/action_language/BESSERActionLanguage.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .BESSERActionLanguageParser import BESSERActionLanguageParser
else:
    from BESSERActionLanguageParser import BESSERActionLanguageParser

# This class defines a complete generic visitor for a parse tree produced by BESSERActionLanguageParser.

class BESSERActionLanguageVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by BESSERActionLanguageParser#function_definition.
    def visitFunction_definition(self, ctx:BESSERActionLanguageParser.Function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#parameter.
    def visitParameter(self, ctx:BESSERActionLanguageParser.ParameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#parameterWithDefault.
    def visitParameterWithDefault(self, ctx:BESSERActionLanguageParser.ParameterWithDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#statements.
    def visitStatements(self, ctx:BESSERActionLanguageParser.StatementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#cond_loop.
    def visitCond_loop(self, ctx:BESSERActionLanguageParser.Cond_loopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#while.
    def visitWhile(self, ctx:BESSERActionLanguageParser.WhileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#do_while.
    def visitDo_while(self, ctx:BESSERActionLanguageParser.Do_whileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#for.
    def visitFor(self, ctx:BESSERActionLanguageParser.ForContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#iterator.
    def visitIterator(self, ctx:BESSERActionLanguageParser.IteratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#conditional_branch.
    def visitConditional_branch(self, ctx:BESSERActionLanguageParser.Conditional_branchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#block.
    def visitBlock(self, ctx:BESSERActionLanguageParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#condition.
    def visitCondition(self, ctx:BESSERActionLanguageParser.ConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#assign_target.
    def visitAssign_target(self, ctx:BESSERActionLanguageParser.Assign_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#explicit_declaration.
    def visitExplicit_declaration(self, ctx:BESSERActionLanguageParser.Explicit_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#assignment.
    def visitAssignment(self, ctx:BESSERActionLanguageParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#return.
    def visitReturn(self, ctx:BESSERActionLanguageParser.ReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#type.
    def visitType(self, ctx:BESSERActionLanguageParser.TypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#single_type.
    def visitSingle_type(self, ctx:BESSERActionLanguageParser.Single_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#sequence_type.
    def visitSequence_type(self, ctx:BESSERActionLanguageParser.Sequence_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#optional_type.
    def visitOptional_type(self, ctx:BESSERActionLanguageParser.Optional_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#function_type.
    def visitFunction_type(self, ctx:BESSERActionLanguageParser.Function_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#any_type.
    def visitAny_type(self, ctx:BESSERActionLanguageParser.Any_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#real_type.
    def visitReal_type(self, ctx:BESSERActionLanguageParser.Real_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#string_type.
    def visitString_type(self, ctx:BESSERActionLanguageParser.String_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#int_type.
    def visitInt_type(self, ctx:BESSERActionLanguageParser.Int_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#bool_type.
    def visitBool_type(self, ctx:BESSERActionLanguageParser.Bool_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#nothing.
    def visitNothing(self, ctx:BESSERActionLanguageParser.NothingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#classifier_type.
    def visitClassifier_type(self, ctx:BESSERActionLanguageParser.Classifier_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#expression.
    def visitExpression(self, ctx:BESSERActionLanguageParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#ternary.
    def visitTernary(self, ctx:BESSERActionLanguageParser.TernaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#boolean.
    def visitBoolean(self, ctx:BESSERActionLanguageParser.BooleanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#or.
    def visitOr(self, ctx:BESSERActionLanguageParser.OrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#and.
    def visitAnd(self, ctx:BESSERActionLanguageParser.AndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#equality.
    def visitEquality(self, ctx:BESSERActionLanguageParser.EqualityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#comparison.
    def visitComparison(self, ctx:BESSERActionLanguageParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#instanceof.
    def visitInstanceof(self, ctx:BESSERActionLanguageParser.InstanceofContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#arithmetic.
    def visitArithmetic(self, ctx:BESSERActionLanguageParser.ArithmeticContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#plus_minus.
    def visitPlus_minus(self, ctx:BESSERActionLanguageParser.Plus_minusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#mult_div.
    def visitMult_div(self, ctx:BESSERActionLanguageParser.Mult_divContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#remain.
    def visitRemain(self, ctx:BESSERActionLanguageParser.RemainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#Cast.
    def visitCast(self, ctx:BESSERActionLanguageParser.CastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#Not.
    def visitNot(self, ctx:BESSERActionLanguageParser.NotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#NullCoalessing.
    def visitNullCoalessing(self, ctx:BESSERActionLanguageParser.NullCoalessingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#Select.
    def visitSelect(self, ctx:BESSERActionLanguageParser.SelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#Minus.
    def visitMinus(self, ctx:BESSERActionLanguageParser.MinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#ArrayAccess.
    def visitArrayAccess(self, ctx:BESSERActionLanguageParser.ArrayAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#FunctionCall.
    def visitFunctionCall(self, ctx:BESSERActionLanguageParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#FieldAccess.
    def visitFieldAccess(self, ctx:BESSERActionLanguageParser.FieldAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#Atom.
    def visitAtom(self, ctx:BESSERActionLanguageParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#field_access.
    def visitField_access(self, ctx:BESSERActionLanguageParser.Field_accessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#array_access.
    def visitArray_access(self, ctx:BESSERActionLanguageParser.Array_accessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#function_call.
    def visitFunction_call(self, ctx:BESSERActionLanguageParser.Function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#atomic.
    def visitAtomic(self, ctx:BESSERActionLanguageParser.AtomicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#procedure_call.
    def visitProcedure_call(self, ctx:BESSERActionLanguageParser.Procedure_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#this.
    def visitThis(self, ctx:BESSERActionLanguageParser.ThisContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#new.
    def visitNew(self, ctx:BESSERActionLanguageParser.NewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#literal.
    def visitLiteral(self, ctx:BESSERActionLanguageParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#single_literal.
    def visitSingle_literal(self, ctx:BESSERActionLanguageParser.Single_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#int_literal.
    def visitInt_literal(self, ctx:BESSERActionLanguageParser.Int_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#string_literal.
    def visitString_literal(self, ctx:BESSERActionLanguageParser.String_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#bool_literal.
    def visitBool_literal(self, ctx:BESSERActionLanguageParser.Bool_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#real_literal.
    def visitReal_literal(self, ctx:BESSERActionLanguageParser.Real_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#null_literal.
    def visitNull_literal(self, ctx:BESSERActionLanguageParser.Null_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#enum_literal.
    def visitEnum_literal(self, ctx:BESSERActionLanguageParser.Enum_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#sequence_literal.
    def visitSequence_literal(self, ctx:BESSERActionLanguageParser.Sequence_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#range_literal.
    def visitRange_literal(self, ctx:BESSERActionLanguageParser.Range_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#symbol.
    def visitSymbol(self, ctx:BESSERActionLanguageParser.SymbolContext):
        return self.visitChildren(ctx)



del BESSERActionLanguageParser


import subprocess
import os
import sys
import tempfile
import shutil
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.generators.sql_alchemy import SQLAlchemyGenerator

class SQLGenerator(GeneratorInterface):
    def __init__(self, model: DomainModel, output_dir: str = None, sql_dialect: str = "sqlite"):
        super().__init__(model, output_dir)
        self.sql_dialect = sql_dialect

    def generate(self):
        """
        Generates SQL code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory is not specified, the code will be stored in the <current directory>/output folder.
        """
        # Determine output path
        final_output_dir = self.output_dir or os.path.join(os.getcwd(), "output")
        os.makedirs(final_output_dir, exist_ok=True)

        # Create temporary workspace
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_py_path = os.path.join(temp_dir, "sql_alchemy.py")

            # Generate SQLAlchemy code into the temporary folder
            alchemy_gen = SQLAlchemyGenerator(model=self.model, output_dir=temp_dir)
            alchemy_gen.generate(dbms=self.sql_dialect)

            # Read and strip final line (optional)
            with open(temp_py_path, "r", encoding="utf-8") as f:
                lines = f.read().splitlines()
            lines = lines[:-1]  # remove last line if needed
            content = "\n".join(lines)
            dialect = self.sql_dialect

            # Append DDL dump logic
            content += f"""

from sqlalchemy.schema import CreateTable
import os

current_directory = os.path.dirname(os.path.realpath(__file__))
sql_file_path = os.path.join(current_directory, f"tables_{dialect}.sql")

ddl_statements = []

# --- Emit ENUM types ---
for table in Base.metadata.tables.values():
    for col in table.columns:
        if isinstance(col.type, Enum):
            enum_name = col.type.name or f"{{col.name}}_enum"
            enum_values = [repr(e.value) for e in col.type.enum_class]
            ddl_statements.append(f"CREATE TYPE {{enum_name}} AS ENUM ({{', '.join(enum_values)}});")

for table in Base.metadata.sorted_tables:
    ddl_statements.append(str(CreateTable(table).compile(engine)))

with open(sql_file_path, "w") as f:
    for stmt in ddl_statements:
        f.write(f"{{stmt}};\\n\\n")
"""

            # Write back modified file
            with open(temp_py_path, "w", encoding="utf-8") as f:
                f.write(content)

            # Run the temporary Python file
            try:
                subprocess.run([sys.executable, temp_py_path], check=True, text=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                print(f"Error running generated SQLAlchemy file:\n{e.stderr}")

            # Copy generated .sql file to final output location
            src_sql = os.path.join(temp_dir, f"tables_{self.sql_dialect}.sql")
            dst_sql = os.path.join(final_output_dir, f"tables_{self.sql_dialect}.sql")
            shutil.copyfile(src_sql, dst_sql)



from besser.BUML.metamodel.structural import DomainModel, MethodImplementationType, Method, Class

from besser.BUML.metamodel.action_language.action_language import *
from besser.BUML.notations.action_language.ActionLanguageASTBuilder import parse_bal


def get_model() -> DomainModel:
    from besser.BUML.metamodel.structural import Property, IntegerType, FloatType, DateType, \
        TimeType, StringType, BinaryAssociation, Multiplicity, Parameter
    # Classes
    Book = Class(name="Book")
    Library = Class(name="Library")
    Author = Class(name="Author")

    # Book class attributes and methods
    Book_title: Property = Property(name="title", type=StringType)
    Book_pages: Property = Property(name="pages", type=IntegerType)
    Book_stock: Property = Property(name="stock", type=IntegerType)
    Book_price: Property = Property(name="price", type=FloatType)
    Book_release: Property = Property(name="release", type=DateType)
    Book_time: Property = Property(name="time", type=TimeType)
    Book_m_decrease_stock: Method = Method(name="decrease_stock", parameters=[Parameter(name='qty', type=IntegerType)], implementation_type=MethodImplementationType.NONE)

    Book.attributes={Book_title, Book_release, Book_pages, Book_stock, Book_time, Book_price}
    Book.methods={Book_m_decrease_stock}

    # Library class attributes and methods
    Library_name: Property = Property(name="name", type=StringType)
    Library.attributes={Library_name}

    # Author class attributes and methods
    Author_name: Property = Property(name="name", type=StringType)
    Author.attributes={Author_name}

    # Relationships
    books: BinaryAssociation = BinaryAssociation(
        name="books",
        ends={
            Property(name="library", type=Library, multiplicity=Multiplicity(0, 9999)),
            Property(name="books", type=Book, multiplicity=Multiplicity(0, 9999))
        }
    )
    books_1: BinaryAssociation = BinaryAssociation(
        name="books_1",
        ends={
            Property(name="authors", type=Author, multiplicity=Multiplicity(0, 9999)),
            Property(name="books", type=Book, multiplicity=Multiplicity(1, 9999))
        }
    )

    return DomainModel(
        name="Class_Diagram",
        types={Book, Library, Author},
        associations={books, books_1},
        generalizations={},
        metadata=None
    )

def test_parse_function_def():
    # setup
    model = get_model()
    library:Class = model.get_type_by_name("Library")
    method: Method = Method(name="method",
                            parameters=[],
                            type=None,
                            implementation_type=MethodImplementationType.BAL
                    )
    method.code = """def method() -> nothing {}"""
    library.add_method(method)

    # parsing
    function:FunctionDefinition = parse_bal(model, library, method.code)

    expected_type = FunctionType([None], Nothing())

    assert function is not None
    assert function.name == "method"
    assert function.return_type == Nothing()
    assert function.parameters[0].name == "self"
    assert function.declared_type == expected_type

def test_parse_types():
    # setup
    model = get_model()
    library:Class = model.get_type_by_name("Library")
    method: Method = Method(name="method",
                            parameters=[],
                            type=None,
                            implementation_type=MethodImplementationType.BAL
                    )
    method.code = """def method() -> nothing {
        var1:bool = null;
        var2:int = null;
        var3:float = null;
        var4:str = null;
        var5:any = null;
        var6:int[] = null;
        var7:Author = null;
        var8:[Book, Library] -> nothing = null;
        var9:[Book?] -> nothing = null;
    }"""
    library.add_method(method)

    # parsing
    function:FunctionDefinition = parse_bal(model, library, method.code)

    expected_type = FunctionType([None], Nothing())

    statements:list[Assignment] = function.body.statements
    targets:list[ExplicitDecl] = [s.target for s in statements]
    types:list[Type] = [t.declared_type for t in targets]
    assert types[0] == BoolType()
    assert types[1] == IntType()
    assert types[2] == RealType()
    assert types[3] == StringType()
    assert types[4] == AnyType()
    assert types[5] == SequenceType(IntType())
    assert types[6] == ObjectType(model.get_type_by_name("Author"))
    var8_type = FunctionType([
                                    ObjectType(model.get_type_by_name("Book")),
                                    ObjectType(model.get_type_by_name("Library"))
                                ],
                                Nothing()
                            )
    var9_type = FunctionType([
                                    OptionalType(ObjectType(model.get_type_by_name("Book")))
                                ],
                                Nothing()
                            )
    assert types[7] == var8_type
    assert types[8] == var9_type


"""
Agent Model Builder

This module generates Python code for BUML agent models.
"""

import os
from re import search
from besser.BUML.metamodel.state_machine.agent import Agent, AgentReply, LLMReply, RAGReply
from besser.BUML.metamodel.state_machine.state_machine import CustomCodeAction


def agent_model_to_code(model: Agent, file_path: str):
    """
    Generates Python code for a B-UML Agent model and writes it to a specified file.

    Parameters:
    model (Agent): The B-UML Agent model object containing states, intents, and transitions.
    file_path (str): The path where the generated code will be saved.

    Outputs:
    - A Python file containing the code representation of the B-UML agent model.
    """
    output_dir = os.path.dirname(file_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not file_path.endswith('.py'):
        file_path += '.py'

    with open(file_path, 'w', encoding='utf-8') as f:
        # Write imports
        f.write("###############\n")
        f.write("# AGENT MODEL #\n")
        f.write("###############\n")
        f.write("import datetime\n")
        f.write(
            "from besser.BUML.metamodel.state_machine.state_machine import "
            "Body, Condition, Event, ConfigProperty, CustomCodeAction\n"
        )
        f.write(
            "from besser.BUML.metamodel.state_machine.agent import "
            "Agent, AgentSession, AgentReply, LLMReply, RAGReply, "
            "LLMOpenAI, LLMHuggingFace, LLMHuggingFaceAPI, LLMReplicate, "
            "RAGVectorStore, RAGTextSplitter\n"
        )
        f.write("from besser.BUML.metamodel.structural import Metadata\n")
        f.write("import operator\n\n")

        # Create agent with metadata if it exists
        if hasattr(model, 'metadata') and model.metadata and model.metadata.description:
            desc = model.metadata.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
            f.write(f"agent = Agent('{model.name}', metadata=Metadata(description=\"{desc}\"))\n\n")
        else:
            f.write(f"agent = Agent('{model.name}')\n\n")
        
        # Write configuration properties
        for prop in model.properties:
            f.write(f"agent.add_property(ConfigProperty('{prop.section}', '{prop.name}', {repr(prop.value)}))\n")
        f.write("\n")

        # Write intents
        f.write("# INTENTS\n")
        for intent in model.intents:
            f.write(f"{intent.name} = agent.new_intent('{intent.name}', [\n")
            for sentence in intent.training_sentences:
                # Escape single quotes for Python string literal
                escaped_sentence = sentence.replace('\\', '\\\\').replace("'", "\\'")
                f.write(f"    '{escaped_sentence}',\n")
            f.write("],\n")
            if intent.description:
                desc = intent.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                f.write(f"description=\"{desc}\"")
            f.write(")\n")
        f.write("\n")

        rag_configs = getattr(model, 'rags', []) or []
        if rag_configs:
            f.write("# RAG CONFIGURATIONS\n")
            for index, rag in enumerate(rag_configs):
                vector_store = getattr(rag, 'vector_store', None)
                splitter = getattr(rag, 'splitter', None)
                if not vector_store or not splitter:
                    continue
                base_name = f"rag_{index}"
                vector_var = f"{base_name}_vector_store"
                splitter_var = f"{base_name}_splitter"
                rag_var = f"{base_name}_rag"

                f.write(f"{vector_var} = RAGVectorStore(\n")
                f.write(f"    embedding_provider={repr(vector_store.embedding_provider)},\n")
                f.write(f"    embedding_parameters={repr(vector_store.embedding_parameters or {})},\n")
                f.write(f"    persist_directory={repr(vector_store.persist_directory)},\n")
                f.write(")\n")

                f.write(f"{splitter_var} = RAGTextSplitter(\n")
                f.write(f"    splitter_type={repr(splitter.splitter_type)},\n")
                f.write(f"    chunk_size={splitter.chunk_size},\n")
                f.write(f"    chunk_overlap={splitter.chunk_overlap},\n")
                f.write(")\n")

                f.write(f"{rag_var} = agent.new_rag(\n")
                f.write(f"    name={repr(rag.name)},\n")
                f.write(f"    vector_store={vector_var},\n")
                f.write(f"    splitter={splitter_var},\n")
                f.write(f"    llm_name={repr(rag.llm_name)},\n")
                f.write(f"    k={rag.k},\n")
                f.write(f"    num_previous_messages={rag.num_previous_messages},\n")
                f.write(")\n\n")
        
        # Check if an LLM is necessary
        llm_required = False
        for state in model.states:
            if state.body:
                if hasattr(state.body, 'actions') and isinstance(state.body.actions[0], LLMReply):
                    llm_required = True
                    break
                    
            if state.fallback_body:
                if hasattr(state.fallback_body, 'actions') and isinstance(state.fallback_body.actions[0], LLMReply):
                    llm_required = True
                    break
        if llm_required:
            # Create an LLM instance for use in state bodies
            f.write("# Create LLM instance for use in state bodies\n")
            f.write("llm = LLMOpenAI(agent=agent, name='gpt-4o-mini', parameters={})\n\n")

        # Write states
        f.write("# STATES\n")
        for state in model.states:
            f.write(f"{state.name} = agent.new_state('{state.name}'")
            if state.initial:
                f.write(", initial=True")
            f.write(")\n")
        f.write("\n")

        # Write state metadata if any states have it
        for state in model.states:
            if hasattr(state, 'metadata') and state.metadata and state.metadata.description:
                desc = state.metadata.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                f.write(f"{state.name}.metadata = Metadata(description=\"{desc}\")\n")
        if any(hasattr(state, 'metadata') and state.metadata and state.metadata.description for state in model.states):
            f.write("\n")

        # Write bodies for states
        for state in model.states:
            f.write(f"# {state.name} state\n")
            # Write body function if it exists
            if state.body:
                # Check if the body has a messages attribute
                if hasattr(state.body, 'actions') and state.body.actions:
                    if isinstance(state.body.actions[0], AgentReply):
                        f.write(f"{state.name}_body = Body('{state.name}_body')\n")
                        for action in state.body.actions:
                            msg = action.message.replace('\\', '\\\\').replace("'", "\\'")
                            f.write(f"{state.name}_body.add_action(AgentReply('{msg}'))\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_body({state.name}_body)\n")
                    elif isinstance(state.body.actions[0], LLMReply):
                        f.write(f"{state.name}_body = Body('{state.name}_body')\n")
                        for action in state.body.actions:
                            prompt = getattr(action, 'prompt', None)
                            if prompt:
                                escaped_prompt = prompt.replace('\\', '\\\\').replace("'", "\\'")
                                f.write(f"{state.name}_body.add_action(LLMReply(prompt='{escaped_prompt}'))\n")
                            else:
                                f.write(f"{state.name}_body.add_action(LLMReply())\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_body({state.name}_body)\n")
                    elif isinstance(state.body.actions[0], RAGReply):
                        f.write(f"{state.name}_body = Body('{state.name}_body')\n")
                        for action in state.body.actions:
                            rag_name = (action.rag_db_name or '').replace('\\', '\\\\').replace("'", "\\'")
                            prompt = getattr(action, 'prompt', None)
                            if prompt:
                                escaped_prompt = prompt.replace('\\', '\\\\').replace("'", "\\'")
                                f.write(
                                    f"{state.name}_body.add_action(RAGReply('{rag_name}', prompt='{escaped_prompt}'))\n"
                                )
                            else:
                                f.write(f"{state.name}_body.add_action(RAGReply('{rag_name}'))\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_body({state.name}_body)\n")
                    elif isinstance(state.body.actions[0], CustomCodeAction):
                        action = state.body.actions[0]
                        f.write(f"{action.to_code()}\n")
                        function_match = search(r'def\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*\(', action.code)
                        f.write(
                            f"CustomCodeAction_{state.name} = "
                            f"CustomCodeAction(callable={function_match.group(1)})\n"
                        )
                        f.write(f"{state.name}_body = Body('{state.name}_body')\n")
                        f.write(f"{state.name}_body.add_action(CustomCodeAction_{state.name})\n")
                        f.write(f"{state.name}.set_body({state.name}_body)\n")

            # Write fallback body function if it exists
            if state.fallback_body:
                # Check if the fallback body has a messages attribute
                if hasattr(state.fallback_body, 'actions') and state.fallback_body.actions:
                    if isinstance(state.fallback_body.actions[0], AgentReply):
                        f.write(f"{state.name}_fallback_body = Body('{state.name}_fallback_body')\n")
                        for action in state.fallback_body.actions:
                            msg = action.message.replace('\\', '\\\\').replace("'", "\\'")
                            f.write(f"{state.name}_fallback_body.add_action(AgentReply('{msg}'))\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_fallback_body({state.name}_fallback_body)\n")
                    elif isinstance(state.fallback_body.actions[0], LLMReply):
                        f.write(f"{state.name}_fallback_body = Body('{state.name}_fallback_body')\n")
                        for action in state.fallback_body.actions:
                            prompt = getattr(action, 'prompt', None)
                            if prompt:
                                escaped_prompt = prompt.replace('\\', '\\\\').replace("'", "\\'")
                                f.write(
                                    f"{state.name}_fallback_body.add_action(LLMReply(prompt='{escaped_prompt}'))\n"
                                )
                            else:
                                f.write(f"{state.name}_fallback_body.add_action(LLMReply())\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_fallback_body({state.name}_fallback_body)\n")
                    elif isinstance(state.fallback_body.actions[0], RAGReply):
                        f.write(f"{state.name}_fallback_body = Body('{state.name}_fallback_body')\n")
                        for action in state.fallback_body.actions:
                            rag_name = (action.rag_db_name or '').replace('\\', '\\\\').replace("'", "\\'")
                            prompt = getattr(action, 'prompt', None)
                            if prompt:
                                escaped_prompt = prompt.replace('\\', '\\\\').replace("'", "\\'")
                                f.write(f"{state.name}_fallback_body.add_action(\n")
                                f.write(f"    RAGReply('{rag_name}', prompt='{escaped_prompt}')\n")
                                f.write(")\n")
                            else:
                                f.write(f"{state.name}_fallback_body.add_action(RAGReply('{rag_name}'))\n")
                        f.write("\n")
                        f.write(f"{state.name}.set_fallback_body({state.name}_fallback_body)\n")
                    elif isinstance(state.fallback_body.actions[0], CustomCodeAction):
                        action = state.fallback_body.actions[0]
                        f.write(f"{action.to_code()}\n")
                        function_match = search(r'def\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*\(', action.code)
                        f.write(
                            f"CustomCodeAction_{state.name}_fallback = "
                            f"CustomCodeAction(callable={function_match.group(1)})\n"
                        )
                        f.write(f"{state.name}_fallback_body = Body('{state.name}_fallback_body')\n")
                        f.write(f"{state.name}_fallback_body.add_action(CustomCodeAction_{state.name}_fallback)\n")
                        f.write(f"{state.name}.set_fallback_body({state.name}_fallback_body)\n")
            
            # Write transitions
            for transition in state.transitions:
                dest_state = transition.dest
                
                # Handle different types of transitions
                if transition.conditions:
                    # Check the type of condition
                    condition_class = transition.conditions.__class__.__name__
                    
                    if condition_class == "IntentMatcher":
                        intent_name = transition.conditions.intent.name
                        if intent_name == "fallback_intent":
                            f.write(f"{state.name}.when_no_intent_matched().go_to({dest_state.name})\n")
                        else:
                            f.write(f"{state.name}.when_intent_matched({intent_name}).go_to({dest_state.name})\n")
                    
                    elif condition_class == "VariableOperationMatcher":
                        var_name = transition.conditions.var_name
                        op_name = transition.conditions.operation.__name__
                        target = transition.conditions.target
                        f.write(f"{state.name}.when_variable_matches_operation(\n")
                        f.write(f"    var_name='{var_name}',\n")
                        f.write(f"    operation=operator.{op_name},\n")
                        f.write(f"    target='{target}'\n")
                        f.write(f").go_to({dest_state.name})\n")
                    
                    elif condition_class == "FileTypeMatcher":
                        file_type = transition.conditions.allowed_types
                        f.write(f"{state.name}.when_file_received('{file_type}').go_to({dest_state.name})\n")
                    
                    elif condition_class == "Auto":
                        f.write(f"{state.name}.go_to({dest_state.name})\n")
                    
                    else:
                        # Default case for custom conditions
                        f.write(f"# Custom transition from {state.name} to {dest_state.name}\n")
                        f.write(f"{state.name}.when_no_intent_matched().go_to({dest_state.name})\n")
                
                else:
                    # If no conditions, create a simple transition
                    f.write(f"{state.name}.go_to({dest_state.name})\n")
                
                f.write("\n")

    print(f"Agent model saved to {file_path}")



from __future__ import annotations

import json
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Set

from besser.BUML.metamodel.gui import (
    Button,
    DataList,
    EmbeddedContent,
    Form,
    GUIModel,
    Image,
    InputField,
    Link,
    Menu,
    MenuItem,
    Text,
    ViewComponent,
    ViewContainer,
)
from besser.BUML.metamodel.gui.dashboard import (
    AgentComponent,
    BarChart,
    FieldColumn,
    LookupColumn,
    ExpressionColumn,
    LineChart,
    MetricCard,
    PieChart,
    RadarChart,
    RadialBarChart,
    Table,
)
from besser.BUML.metamodel.gui.events_actions import (
    Create,
    Delete,
    Event,
    Parameter,
    Read,
    Transition,
    Update,
)
from besser.BUML.metamodel.structural import Class, DomainModel, Enumeration
from besser.utilities import sort_by_timestamp


class GuiSerializationMixin:
    def _build_generation_context(self) -> Dict[str, Any]:
        components_payload, styles_payload, meta = self._serialize_gui_model()
        
        # Collect all unique field names from all FieldColumns across all Tables
        all_table_fields = self._collect_all_table_fields()
        
        # Get used component types for conditional rendering in templates
        used_component_types = self._get_used_component_types()
        
        return {
            "model": self.gui_model,
            "components_json": self._to_pretty_json(components_payload),
            "styles_json": self._to_pretty_json(styles_payload),
            "main_page_id": meta.get("main_page_id"),
            "main_page_name": meta.get("main_page_name"),
            "pages_meta": meta.get("pages", []),
            "module_name": meta.get("module_name"),
            "all_table_fields": all_table_fields,
            "used_components": used_component_types,
        }

    def _serialize_gui_model(self) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        pages: List[Dict[str, Any]] = []
        pages_meta: List[Dict[str, str]] = []
        self._style_map = {}

        modules = self._sorted_by_name(self.gui_model.modules)
        module_name = modules[0].name if modules else "AppModule"
        main_page_id: Optional[str] = None
        main_page_name: Optional[str] = None

        for module in modules:
            for screen in self._sorted_by_name(module.screens):
                screen_node = self._serialize_screen(screen)
                pages.append(screen_node)
                pages_meta.append(
                    {
                        "id": screen_node["id"],
                        "name": screen_node.get("name") or screen_node["id"],
                        "route_path": screen_node.get("route_path"),
                    }
                )
                if screen_node.get("is_main") and main_page_id is None:
                    main_page_id = screen_node["id"]
                    main_page_name = screen_node.get("name")

        if main_page_id is None and pages:
            main_page_id = pages[0]["id"]
            main_page_name = pages[0].get("name")

        # Merge raw style entries from the modelling editor LAST
        # GrapesJS styles should COMPLETELY override BUML-generated styles
        for entry in self._raw_style_entries:
            selectors = entry.get("selectors") or []
            grapesjs_style = self._convert_style_keys(entry.get("style") or {})
            
            # MODERNIZE GRAPESJS GRID SYSTEM: Convert table-based grid to flexbox
            if ".gjs-row" in selectors or "gjs-row" in selectors:
                # Convert table display to flex for better responsiveness
                if grapesjs_style.get("display") == "table":
                    grapesjs_style["display"] = "flex"
                    grapesjs_style["flexWrap"] = "wrap"  # Allow wrapping for responsive layout
            
            if ".gjs-cell" in selectors or "gjs-cell" in selectors:
                # Convert table-cell to flex item with proper sizing for 3-column responsive layout
                if grapesjs_style.get("display") in ("table-cell", "block"):
                    # Don't force display for cells - let them be flex items
                    if "display" in grapesjs_style:
                        del grapesjs_style["display"]
                
                # Always use flex-basis for gjs-cell to get proper 3-column layout
                # GrapesJS uses outdated table widths (8%), modernize to flex
                grapesjs_style["flex"] = "1 1 calc(33.333% - 20px)"  # 3 columns with gap
                grapesjs_style["minWidth"] = "250px"  # Min width for responsive wrapping
                # Remove width if present (table-based sizing)
                if "width" in grapesjs_style:
                    del grapesjs_style["width"]
                # Remove height if present (fixed heights break responsiveness)
                if "height" in grapesjs_style:
                    del grapesjs_style["height"]
            
            if selectors:
                # Convert tuple for lookup
                selector_tuple = tuple(selectors)
                # Check if entry already exists from BUML
                existing = self._style_map.get(selector_tuple)
                if existing:
                    # GrapesJS style should override existing BUML style completely
                    # Keep only chart custom properties from BUML, replace everything else
                    buml_chart_props = {k: v for k, v in existing["style"].items() 
                                       if k.startswith("--chart-")}
                    # Start with GrapesJS style, then add back chart props if not overridden
                    merged = dict(grapesjs_style)
                    for chart_key, chart_val in buml_chart_props.items():
                        if chart_key not in merged:
                            merged[chart_key] = chart_val
                    existing["style"] = merged
                else:
                    # New entry - just add it
                    self._add_style_entry(selector_tuple, grapesjs_style)

        styles_payload = {"styles": list(self._style_map.values())}
        components_payload = {"pages": pages}
        meta = {
            "main_page_id": main_page_id,
            "main_page_name": main_page_name,
            "pages": pages_meta,
            "module_name": module_name,
        }
        return components_payload, styles_payload, meta

    # --------------------------------------------------------------------- #
    # Screen & component serialization
    # --------------------------------------------------------------------- #
    def _serialize_screen(self, screen: ViewContainer) -> Dict[str, Any]:
        # Screens may have page_id (preferred) or component_id, fallback to name
        # Note: page_id is set by processor.py for GrapesJS pages
        screen_id = getattr(screen, 'page_id', None) or screen.component_id or screen.name
        screen_name = screen.name
        
        node: Dict[str, Any] = {
            "id": screen_id,
            "name": screen.description or self._humanize(screen_name),
            "description": screen.description or "",
            "is_main": bool(screen.is_main_page),
            "route_path": screen.route_path or f"/{screen_name}".lower().replace(' ', '-'),
            "components": [],
        }

        self._register_component_style(screen_id, screen.styling, getattr(screen, "layout", None))

        # Sort elements by display_order to preserve JSON ordering
        elements = list(screen.view_elements)
        elements.sort(key=lambda e: (e.display_order, e.name))
        
        for element in elements:
            node["components"].append(self._serialize_component(element))

        return self._clean_dict(node)

    def _serialize_component(self, element: ViewComponent) -> Dict[str, Any]:
        component_type = self._map_component_type(element)

        # Use metadata from BUML metamodel (no getattr needed - these are real attributes!)
        # Fallback to defaults only if None
        component_id = element.component_id or element.name or f"{element.__class__.__name__}"
        tag = element.tag_name
        class_list = element.css_classes
        attributes = element.custom_attributes
        display_order = element.display_order

        node: Dict[str, Any] = {
            "id": component_id,
            "type": component_type,
            "name": self._derive_display_name(element),
            "description": element.description or "",
            "tag": tag,
            "class_list": class_list,
            "attributes": attributes,
            "display_order": display_order,
        }

        layout = getattr(element, "layout", None)
        self._register_component_style(component_id, element.styling, layout)

        if isinstance(element, ViewContainer):
            # Sort children by display_order to preserve JSON ordering
            children_list = list(element.view_elements)
            children_list.sort(key=lambda e: (e.display_order, e.name))
            
            children = [self._serialize_component(child) for child in children_list]
            if children:
                node["children"] = children

        if isinstance(element, Text):
            # Ensure content is properly extracted and not empty
            content = element.content
            # If content is empty but we have a tag like h1, h2, h3, p, try to derive from name
            if not content and tag and tag.lower() in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'span', 'div']:
                # Use name or description as fallback
                content = self._derive_display_name(element) or element.description or ""
            node["content"] = content or ""

        if isinstance(element, Image):
            node["alt"] = element.description or ""
            node["src"] = getattr(element, "source", None)

        if isinstance(element, Link):
            node["label"] = element.label
            node["url"] = element.url
            node["target"] = element.target
            node["rel"] = element.rel

        if isinstance(element, EmbeddedContent):
            node["src"] = element.source
            node["content_type"] = element.content_type or element.name
            node["extra_props"] = element.extra_props or None

        if isinstance(element, Button):
            node["label"] = element.label
            node["button_type"] = self._enum_value(getattr(element, "buttonType", None))
            node["action_type"] = self._enum_value(getattr(element, "actionType", None))
            target_screen = getattr(element, "targetScreen", None)
            if target_screen:
                node["target_screen"] = getattr(target_screen, "name", None)
                node["target_screen_path"] = getattr(target_screen, "route_path", None)
            events = self._serialize_events(element)
            if events:
                node["events"] = events

            # Handle method execution using method_btn (Method object reference)
            method_btn = getattr(element, "method_btn", None)
            if method_btn:
                # Get method name and class from the Method object
                method_name = method_btn.name
                method_class = method_btn.owner
                class_name = method_class.name if method_class else None

                if method_name and class_name:
                    # Clean the method name (remove visibility and parameters if present)
                    clean_method_name = method_name.split('(')[0] if '(' in method_name else method_name

                    # Update attributes with clean names (not IDs)
                    if 'data-method-name' in attributes:
                        attributes['method-name'] = clean_method_name
                        del attributes['data-method-name']

                    if 'data-method-class' in attributes:
                        attributes['method-class'] = class_name
                        del attributes['data-method-class']

                    # Generate endpoint URL for method execution
                    # Format: /book/{book_id}/methods/decrease_stock/
                    class_name_lower = class_name.lower()
                    endpoint = f"/{class_name_lower}/{{{class_name_lower}_id}}/methods/{clean_method_name}/"
                    attributes['endpoint'] = endpoint

                    # Use is_instance_method from Button object
                    is_instance = getattr(element, "is_instance_method", False)
                    attributes['is-instance-method'] = str(is_instance).lower()

                    # Extract input parameters from the method
                    if hasattr(method_btn, 'parameters') and method_btn.parameters:
                        input_params = {}
                        for param in method_btn.parameters:
                            # Skip 'self' and 'session' parameters
                            if param.name.lower() not in ('self', 'session'):
                                param_type = getattr(param, "type", None)
                                param_type_name = getattr(param_type, "name", "any")
                                default_value = getattr(param, "default_value", None)

                                param_data: Dict[str, Any] = {
                                    "type": param_type_name,
                                    "required": default_value is None,
                                }

                                if default_value is not None:
                                    if isinstance(default_value, (str, int, float, bool)):
                                        param_data["default"] = default_value
                                    else:
                                        param_data["default"] = str(default_value)

                                if isinstance(param_type, Class):
                                    target_attrs = sort_by_timestamp(list(param_type.all_attributes()))
                                    lookup_field = self._select_lookup_field(target_attrs)
                                    param_data["input_kind"] = "lookup"
                                    param_data["entity"] = param_type_name
                                    if lookup_field:
                                        param_data["lookup_field"] = lookup_field
                                elif isinstance(param_type, Enumeration):
                                    param_data["input_kind"] = "enum"
                                    param_data["options"] = sorted([literal.name for literal in param_type.literals])

                                input_params[param.name] = param_data
                        
                        if input_params:
                            attributes['input-parameters'] = input_params

            # Output instance_source (table/component ID providing instance data)
            instance_source = getattr(element, "instance_source", None)
            if instance_source:
                # If it's a ViewComponent object, get its component_id; otherwise use the string ID
                if hasattr(instance_source, 'component_id'):
                    attributes['instance-source'] = instance_source.component_id
                elif isinstance(instance_source, str):
                    attributes['instance-source'] = instance_source

        if isinstance(element, InputField):
            node["input_type"] = self._enum_value(getattr(element, "field_type", None))
            node["validation"] = getattr(element, "validationRules", None)

        if isinstance(element, Form):
            inputs = []
            for input_field in sorted(
                getattr(element, "inputFields", []), key=lambda f: getattr(f, "name", "").lower()
            ):
                inputs.append(
                    self._clean_dict(
                        {
                            "id": getattr(input_field, "name", None),
                            "label": self._humanize(getattr(input_field, "name", "")),
                            "type": self._enum_value(getattr(input_field, "field_type", None)),
                            "validation": getattr(input_field, "validationRules", None),
                        }
                    )
                )
            if inputs:
                node["inputs"] = inputs

        if isinstance(element, Menu):
            items = []
            for item in self._sorted_menu_items(element.menuItems):
                items.append(
                    self._clean_dict(
                        {
                            "label": item.label,
                            "url": getattr(item, "url", None),
                            "target": getattr(item, "target", None),
                            "rel": getattr(item, "rel", None),
                        }
                    )
                )
            if items:
                node["items"] = items

        if isinstance(element, DataList):
            sources = []
            for source in self._sorted_by_name(element.list_sources):
                domain_name = getattr(getattr(source, "dataSourceClass", None), "name", None)
                field_names = [
                    getattr(field, "name", None)
                    for field in getattr(source, "fields", [])
                    if getattr(field, "name", None)
                ]
                if not field_names and getattr(source, "field_names", None):
                    field_names = list(getattr(source, "field_names"))
                fields = sorted(field_names) if field_names else None
                label_field_name = getattr(getattr(source, "label_field", None), "name", None) or getattr(
                    source, "label_field_name", None
                )
                value_field_name = getattr(getattr(source, "value_field", None), "name", None) or getattr(
                    source, "value_field_name", None
                )
                sources.append(
                    self._clean_dict(
                        {
                            "name": source.name,
                            "domain": domain_name,
                            "fields": fields,
                            "label_field": label_field_name,
                            "value_field": value_field_name,
                        }
                    )
                )
            if sources:
                node["data_sources"] = sources

        if isinstance(element, LineChart):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "lineWidth": element.line_width,
                    "curveType": element.curve_type,
                    "showGrid": element.show_grid,
                    "showLegend": element.show_legend,
                    "showTooltip": element.show_tooltip,
                    "animate": element.animate,
                    "legendPosition": element.legend_position,
                    "gridColor": element.grid_color,
                    "dotSize": element.dot_size,
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("line") or "#4CAF50"
            # Serialize series with data bindings
            series_list = self._serialize_chart_series(element.series)
            if series_list:
                node["series"] = series_list

        if isinstance(element, BarChart):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "barWidth": element.bar_width,
                    "orientation": element.orientation,
                    "showGrid": element.show_grid,
                    "showLegend": element.show_legend,
                    "showTooltip": element.show_tooltip,
                    "stacked": element.stacked,
                    "animate": element.animate,
                    "legendPosition": element.legend_position,
                    "gridColor": element.grid_color,
                    "barGap": element.bar_gap,
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("bar") or "#3498db"
            # Serialize series with data bindings
            series_list = self._serialize_chart_series(element.series)
            if series_list:
                node["series"] = series_list

        if isinstance(element, PieChart):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "showLegend": element.show_legend,
                    "legendPosition": self._enum_value(element.legend_position),
                    "showLabels": element.show_labels,
                    "labelPosition": self._enum_value(element.label_position),
                    "paddingAngle": element.padding_angle,
                    "innerRadius": element.inner_radius,
                    "outerRadius": element.outer_radius,
                    "startAngle": element.start_angle,
                    "endAngle": element.end_angle,
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("palette") or "#8884d8"
            # Serialize series with data bindings
            series_list = self._serialize_chart_series(element.series)
            if series_list:
                node["series"] = series_list

        if isinstance(element, RadarChart):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "showGrid": element.show_grid,
                    "showTooltip": element.show_tooltip,
                    "showLegend": element.show_legend,
                    "legendPosition": element.legend_position,
                    "dotSize": element.dot_size,
                    "gridType": element.grid_type,
                    "strokeWidth": element.stroke_width,
                    "showRadiusAxis": element.show_radius_axis,
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("palette") or "#8884d8"
            # Serialize series with data bindings
            series_list = self._serialize_chart_series(element.series)
            if series_list:
                node["series"] = series_list

        if isinstance(element, RadialBarChart):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "startAngle": element.start_angle,
                    "endAngle": element.end_angle,
                    "innerRadius": element.inner_radius,
                    "outerRadius": element.outer_radius,
                    "showLegend": element.show_legend,
                    "legendPosition": element.legend_position,
                    "showTooltip": element.show_tooltip,
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("palette") or "#8884d8"
            # Serialize series with data bindings
            series_list = self._serialize_chart_series(element.series)
            if series_list:
                node["series"] = series_list

        if isinstance(element, Table):
            node["title"] = element.title or self._humanize(element.name)
            node["chart"] = self._clean_dict(
                {
                    "showHeader": element.show_header,
                    "stripedRows": element.striped_rows,
                    "showPagination": element.show_pagination,
                    "rowsPerPage": element.rows_per_page,
                    "actionButtons": getattr(element, "action_buttons", False),
                }
            )
            chart_colors = self._extract_chart_colors(element)
            node["color"] = element.primary_color or chart_colors.get("background") or "#2c3e50"

            # Serialize columns from Column objects
            columns = []
            for col in getattr(element, "columns", []) or []:
                column_dict = {"label": col.label}

                if isinstance(col, FieldColumn):
                    column_dict["column_type"] = "field"
                    column_dict["field"] = col.field.name if hasattr(col.field, "name") else str(col.field)

                    # Check if the field type is an Enumeration
                    field_type = getattr(col.field, "type", None)
                    if isinstance(field_type, Enumeration):
                        column_dict["type"] = "enum"
                        # Extract enumeration literal names as options
                        column_dict["options"] = sorted([literal.name for literal in field_type.literals])
                    else:
                        column_dict["type"] = field_type.name if hasattr(field_type, "name") else "str"
                    
                    # Check if the field is required based on minimum multiplicity
                    if hasattr(col.field, "multiplicity") and hasattr(col.field.multiplicity, "min"):
                        column_dict["required"] = col.field.multiplicity.min > 0
                    else:
                        column_dict["required"] = False

                elif isinstance(col, LookupColumn):
                    column_dict["column_type"] = "lookup"
                    column_dict["path"] = col.path.name if hasattr(col.path, "name") else str(col.path)
                    column_dict["entity"] = col.path.type.name if hasattr(col.path, "type") and hasattr(col.path.type, "name") else ""
                    column_dict["field"] = col.field.name if hasattr(col.field, "name") else str(col.field)
                    # Determine type based on path multiplicity
                    if hasattr(col.path, "multiplicity") and hasattr(col.path.multiplicity, "max"):
                        is_list = col.path.multiplicity.max > 1 or col.path.multiplicity.max == "*"
                        if is_list:
                            column_dict["type"] = "list"
                        else:
                            column_dict["type"] = getattr(col.field, "type", {}).name if hasattr(getattr(col.field, "type", None), "name") else "str"
                    else:
                        column_dict["type"] = getattr(col.field, "type", {}).name if hasattr(getattr(col.field, "type", None), "name") else "str"
                    
                    # Check if the lookup path is required based on minimum multiplicity
                    if hasattr(col.path, "multiplicity") and hasattr(col.path.multiplicity, "min"):
                        column_dict["required"] = col.path.multiplicity.min > 0
                    else:
                        column_dict["required"] = False

                elif isinstance(col, ExpressionColumn):
                    column_dict["column_type"] = "expression"
                    column_dict["field"] = col.label
                    column_dict["expression"] = col.expression
                    column_dict["type"] = "computed"

                columns.append(column_dict)

            if columns:
                node["chart"]["columns"] = columns

            # Build complete form columns from domain metadata (all attributes + associations)
            form_columns = []
            binding = getattr(element, "data_binding", None)
            domain_concept = getattr(binding, "domain_concept", None) if binding else None
            if domain_concept:
                attributes = list(domain_concept.all_attributes())
                attributes = sort_by_timestamp(attributes) if attributes else []
                for attr in attributes:
                    field_type = getattr(attr, "type", None)
                    column_dict = {
                        "column_type": "field",
                        "field": attr.name,
                        "label": attr.name,
                        "type": "str",
                        "required": False,
                        "defaultValue": getattr(attr, "default_value", None),
                    }

                    if isinstance(field_type, Enumeration):
                        column_dict["type"] = "enum"
                        column_dict["options"] = sorted([literal.name for literal in field_type.literals])
                    elif field_type and getattr(field_type, "name", None):
                        column_dict["type"] = field_type.name

                    is_optional = getattr(attr, "is_optional", False)
                    if not is_optional:
                        multiplicity = getattr(attr, "multiplicity", None)
                        if multiplicity and getattr(multiplicity, "min", 0) > 0:
                            column_dict["required"] = True

                    form_columns.append(column_dict)

                ends = list(domain_concept.all_association_ends())
                ends = sort_by_timestamp(ends) if ends else []
                for end in ends:
                    if hasattr(end, "is_navigable") and not end.is_navigable:
                        continue

                    target = getattr(end, "type", None)
                    target_attrs = []
                    if target and hasattr(target, "all_attributes"):
                        target_attrs = sort_by_timestamp(list(target.all_attributes()))
                    lookup_field = target_attrs[0].name if target_attrs else ""

                    max_mult = getattr(getattr(end, "multiplicity", None), "max", None)
                    is_list = max_mult == "*" or (isinstance(max_mult, int) and max_mult > 1)

                    column_dict = {
                        "column_type": "lookup",
                        "path": end.name,
                        "field": end.name,
                        "lookup_field": lookup_field,
                        "entity": getattr(target, "name", "") if target else "",
                        "type": "list" if is_list else "str",
                        "required": False,
                    }

                    multiplicity = getattr(end, "multiplicity", None)
                    if multiplicity and getattr(multiplicity, "min", 0) > 0:
                        column_dict["required"] = True

                    form_columns.append(column_dict)

            if form_columns:
                node["chart"]["formColumns"] = form_columns
            
            # Remove raw GrapesJS columns from attributes - we use the processed columns in chart
            if "attributes" in node and isinstance(node["attributes"], dict) and "columns" in node["attributes"]:
                node["attributes"] = {k: v for k, v in node["attributes"].items() if k != "columns"}

        if isinstance(element, MetricCard):
            node["title"] = element.metric_title or self._humanize(element.name)
            node["metric"] = self._clean_dict(
                {
                    "metricTitle": element.metric_title,
                    "format": element.format,
                    "valueColor": element.value_color,
                    "valueSize": element.value_size,
                    "showTrend": element.show_trend,
                    "positiveColor": element.positive_color,
                    "negativeColor": element.negative_color,
                    "value": 0,  # Default placeholder - will be populated from API
                    "trend": 12,  # Default placeholder trend percentage
                }
            )
            node["color"] = element.primary_color or element.value_color or "#2c3e50"

        if isinstance(element, AgentComponent):
            node["agent-name"] = element.agent_name or ""
            node["agent-title"] = element.agent_title or "BESSER Agent"
            node["agent_name"] = element.agent_name or ""
            node["agent_title"] = element.agent_title or "BESSER Agent"

        # Clean up raw GrapesJS series from attributes for chart components
        # We use the processed series array instead of the raw JSON string
        if isinstance(element, (LineChart, BarChart, PieChart, RadarChart, RadialBarChart)):
            if "attributes" in node and isinstance(node["attributes"], dict) and "series" in node["attributes"]:
                node["attributes"] = {k: v for k, v in node["attributes"].items() if k != "series"}

        binding_data = self._serialize_data_binding(getattr(element, "data_binding", None))
        if binding_data:
            node["data_binding"] = binding_data

        return self._clean_dict(node)

    def _serialize_events(self, element: Button) -> Optional[List[Dict[str, Any]]]:
        events = getattr(element, "events", None)
        if not events:
            return None

        serialized: List[Dict[str, Any]] = []
        for event in self._sorted_by_name(events):
            actions = [
                action_data
                for action in self._sorted_by_name(event.actions)
                if (action_data := self._serialize_action(action))
            ]
            event_dict = self._clean_dict(
                {
                    "name": event.name,
                    "type": self._enum_value(getattr(event, "event_type", None)),
                    "actions": actions,
                }
            )
            if event_dict.get("actions"):
                serialized.append(event_dict)
        return serialized or None

    def _serialize_action(self, action) -> Optional[Dict[str, Any]]:
        if action is None:
            return None

        data: Dict[str, Any] = {
            "name": getattr(action, "name", None),
            "kind": action.__class__.__name__,
            "description": getattr(action, "description", None),
        }

        if isinstance(action, Transition):
            target_screen = getattr(action, "target_screen", None)
            data["target_screen"] = getattr(target_screen, "name", None)
            data["target_screen_id"] = getattr(action, "_target_screen_id", None)
            data["target_screen_path"] = getattr(target_screen, "route_path", None)

        if isinstance(action, (Create, Read, Update, Delete)):
            data["target_class"] = getattr(getattr(action, "target_class", None), "name", None)

        parameters = getattr(action, "parameters", None)
        if parameters:
            params_serialized = []
            for param in sorted(parameters, key=lambda p: getattr(p, "name", "")):
                params_serialized.append(
                    self._clean_dict(
                        {
                            "name": getattr(param, "name", None),
                            "type": getattr(param, "param_type", None),
                            "required": getattr(param, "required", None),
                            "value": getattr(param, "value", None),
                        }
                    )
                )
            if params_serialized:
                data["parameters"] = params_serialized

        return self._clean_dict(data)

    def _serialize_data_binding(self, binding) -> Optional[Dict[str, Any]]:
        if not binding:
            return None

        domain = getattr(binding, "domain_concept", None)
        label_field = getattr(binding, "label_field", None)
        data_field = getattr(binding, "data_field", None)
        data_filter = getattr(binding, "data_filter", None)
        label_field_path = getattr(binding, "label_field_path", None)
        data_field_path = getattr(binding, "data_field_path", None)
        
        endpoint = None
        if domain and getattr(domain, "name", None):
            endpoint = f"/{domain.name.lower()}/"

        # Use path for nested fields, otherwise use field attribute name
        label_field_value = label_field_path if label_field_path else (getattr(label_field, "name", None) if label_field else None)
        data_field_value = data_field_path if data_field_path else (getattr(data_field, "name", None) if data_field else None)

        return self._clean_dict(
            {
                "entity": getattr(domain, "name", None),
                "endpoint": endpoint,
                "label_field": label_field_value,
                "data_field": data_field_value,
                "filter": str(data_filter) if data_filter else None,
            }
        )

    def _serialize_chart_series(self, series_list) -> Optional[List[Dict[str, Any]]]:
        """Serialize chart series with their data bindings."""
        if not series_list:
            return None
        
        serialized = []
        default_colors = ["#4CAF50", "#2196F3", "#FF9800", "#E91E63", "#9C27B0", "#00BCD4", "#FF5722", "#795548"]
        
        for idx, series in enumerate(series_list):
            binding = getattr(series, "data_binding", None)
            styling = getattr(series, "styling", None)
            
            # Extract color from styling or use default
            # Check multiple color properties in order of preference
            color = None
            if styling:
                color_obj = getattr(styling, "color", None)
                if color_obj:
                    # Try primary_color first (set by chart parser), then line_color, bar_color, background_color
                    color = (getattr(color_obj, "primary_color", None) or 
                             getattr(color_obj, "line_color", None) or 
                             getattr(color_obj, "bar_color", None) or
                             getattr(color_obj, "background_color", None))
            if not color:
                color = default_colors[idx % len(default_colors)]
            
            series_data: Dict[str, Any] = {
                "name": series.name,
                "label": series.label or series.name,
                "color": color,
            }
            
            # Add data binding info if available
            if binding:
                domain = getattr(binding, "domain_concept", None)
                label_field = getattr(binding, "label_field", None)
                data_field = getattr(binding, "data_field", None)
                filter_expression = getattr(binding, "filter_expression", None)
                
                if domain:
                    series_data["dataSource"] = domain.name.lower()
                    series_data["endpoint"] = f"/{domain.name.lower()}/"
                
                # Use label_field_path for nested fields, otherwise use label_field attribute
                label_field_path = getattr(binding, "label_field_path", None)
                if label_field_path:
                    series_data["labelField"] = label_field_path
                elif label_field:
                    series_data["labelField"] = getattr(label_field, "name", None)
                
                # Use data_field_path for nested fields, otherwise use data_field attribute
                data_field_path = getattr(binding, "data_field_path", None)
                if data_field_path:
                    series_data["dataField"] = data_field_path
                elif data_field:
                    series_data["dataField"] = getattr(data_field, "name", None)
                
                if filter_expression:
                    series_data["filter"] = str(filter_expression)
            
            serialized.append(self._clean_dict(series_data))
        
        return serialized if serialized else None

    # --------------------------------------------------------------------- #
    # Styling helpers
    # --------------------------------------------------------------------- #
    def _register_component_style(self, selector_id: str, styling, layout=None):
        style = self._style_from_styling(styling, layout)
        # Only register if there are meaningful styles (not just defaults)
        if style and self._has_meaningful_styles(style):
            self._add_style_entry((f"#{selector_id}",), style)

    def _style_from_styling(self, styling, layout=None) -> Dict[str, Any]:
        if not styling:
            return {}

        style: Dict[str, Any] = {}
        size = getattr(styling, "size", None)
        position = getattr(styling, "position", None)
        color = getattr(styling, "color", None)

        if size:
            style.update(self._extract_size_style(size))
        if position:
            style.update(self._extract_position_style(position))
        if color:
            style.update(self._extract_color_style(color))

        if layout is None and hasattr(styling, "layout"):
            layout = styling.layout
        if layout:
            style.update(self._extract_layout_style(layout))

        return self._filter_style(style)

    def _extract_size_style(self, size) -> Dict[str, Any]:
        style: Dict[str, Any] = {}
        if getattr(size, "width", None):
            style["width"] = size.width
        if getattr(size, "height", None):
            style["height"] = size.height
        if getattr(size, "padding", None):
            style["padding"] = size.padding
        if getattr(size, "margin", None):
            style["margin"] = size.margin
        if getattr(size, "font_size", None):
            style["fontSize"] = size.font_size
        if getattr(size, "line_height", None):
            style["lineHeight"] = size.line_height
        if getattr(size, "icon_size", None) and "fontSize" not in style:
            style["fontSize"] = size.icon_size
        return style

    def _extract_position_style(self, position) -> Dict[str, Any]:
        style: Dict[str, Any] = {}
        position_type = getattr(position, "p_type", None)
        if position_type:
            style["position"] = self._enum_value(position_type)

        for attr in ("top", "left", "right", "bottom"):
            value = getattr(position, attr, None)
            if value not in (None, "auto"):
                style[attr] = value

        alignment = getattr(position, "alignment", None)
        if alignment:
            style["textAlign"] = self._enum_value(alignment) or alignment

        if getattr(position, "z_index", None) is not None:
            style["zIndex"] = position.z_index

        return style

    def _extract_color_style(self, color) -> Dict[str, Any]:
        style: Dict[str, Any] = {}
        if getattr(color, "background_color", None):
            bg_color = str(color.background_color).replace(" !important", "").replace("!important", "").strip()
            if "linear-gradient" in bg_color or "radial-gradient" in bg_color:
                style["backgroundImage"] = bg_color
            else:
                style["backgroundColor"] = bg_color
        if getattr(color, "text_color", None):
            style["color"] = str(color.text_color).replace(" !important", "").replace("!important", "").strip()
        if getattr(color, "border_color", None):
            style["borderColor"] = str(color.border_color).replace(" !important", "").replace("!important", "").strip()
        if getattr(color, "opacity", None) not in (None, ""):
            style["opacity"] = color.opacity
        if getattr(color, "box_shadow", None):
            style["boxShadow"] = color.box_shadow
        if getattr(color, "gradient", None):
            style["backgroundImage"] = color.gradient
        # Preserve chart colors as CSS custom properties for styling fallbacks
        if getattr(color, "line_color", None):
            style["--chart-line-color"] = color.line_color
        if getattr(color, "bar_color", None):
            style["--chart-bar-color"] = color.bar_color
        if getattr(color, "color_palette", None):
            style["--chart-color-palette"] = color.color_palette
        if getattr(color, "radius", None):
            style["borderRadius"] = color.radius
        return style

    def _extract_layout_style(self, layout) -> Dict[str, Any]:
        style: Dict[str, Any] = {}

        layout_type = self._enum_value(getattr(layout, "layout_type", None))
        orientation = getattr(layout, "orientation", None)

        if layout_type in {"flex", "row", "column"}:
            style["display"] = "flex"
        elif layout_type == "grid":
            style["display"] = "grid"
        elif layout_type == "absolute":
            style["position"] = "absolute"

        flex_direction = getattr(layout, "flex_direction", None)
        if flex_direction:
            # Only add flexDirection if explicitly set in GrapesJS
            style["flexDirection"] = flex_direction
        elif orientation == "horizontal":
            style["flexDirection"] = "row"
        elif orientation == "vertical":
            style["flexDirection"] = "column"
        elif layout_type == "row":
            style["flexDirection"] = "row"
        elif layout_type == "column":
            style["flexDirection"] = "column"
        # DON'T add default flexDirection - let CSS handle it (default is row anyway)

        if getattr(layout, "justify_content", None):
            style["justifyContent"] = layout.justify_content
        elif getattr(layout, "alignment", None):
            alignment = self._enum_value(layout.alignment)
            if alignment:
                style.setdefault("justifyContent", alignment)

        if getattr(layout, "align_items", None):
            style["alignItems"] = layout.align_items

        if getattr(layout, "flex_wrap", None):
            style["flexWrap"] = layout.flex_wrap
        elif getattr(layout, "wrap", None) is not None:
            style["flexWrap"] = "wrap" if layout.wrap else "nowrap"

        if getattr(layout, "gap", None):
            style["gap"] = layout.gap
        if getattr(layout, "grid_template_columns", None):
            style["gridTemplateColumns"] = layout.grid_template_columns
        if getattr(layout, "grid_template_rows", None):
            style["gridTemplateRows"] = layout.grid_template_rows
        if getattr(layout, "grid_gap", None):
            style["gridGap"] = layout.grid_gap

        if getattr(layout, "padding", None):
            style.setdefault("padding", layout.padding)
        if getattr(layout, "margin", None):
            style.setdefault("margin", layout.margin)

        return style

    def _extract_chart_colors(self, element: ViewComponent) -> Dict[str, Any]:
        color = getattr(getattr(element, "styling", None), "color", None)
        if not color:
            return {}
        return {
            "line": getattr(color, "line_color", None),
            "bar": getattr(color, "bar_color", None),
            "palette": getattr(color, "color_palette", None),
            "background": getattr(color, "background_color", None),
        }

    def _add_style_entry(self, selectors: Sequence[str], style: Dict[str, Any]):
        normalized_selectors = tuple(sorted({selector for selector in selectors if selector}))
        if not normalized_selectors or not style:
            return
        entry = self._style_map.get(normalized_selectors)
        if entry:
            entry["style"].update(style)
        else:
            self._style_map[normalized_selectors] = {
                "selectors": list(normalized_selectors),
                "style": dict(style),
            }

    def _has_meaningful_styles(self, style: Dict[str, Any]) -> bool:
        """Check if the style dict has non-default meaningful values."""
        # These are default values that don't need to be exported
        defaults = {
            "backgroundColor": "#FFFFFF",
            "color": "#000000",
            "borderColor": "#CCCCCC",
            "margin": "0",
            "padding": "0",
            "height": "auto",
            "width": "auto",
            "position": "static",
            "textAlign": "left",
            "zIndex": 0,
            "--chart-line-color": "#000000",
            "--chart-bar-color": "#CCCCCC",
            "--chart-color-palette": "default",
        }
        
        # Check if there's at least one non-default value
        for key, value in style.items():
            if key not in defaults:
                # This is a non-default property
                return True
            if defaults.get(key) != value:
                # This is a default property but with a different value
                return True
        
        # All values are defaults
        return False

    # --------------------------------------------------------------------- #
    # Utility helpers
    # --------------------------------------------------------------------- #
    @staticmethod
    def _to_pretty_json(payload: Dict[str, Any]) -> str:
        return json.dumps(payload, indent=2, sort_keys=True, ensure_ascii=False)

    @staticmethod
    def _select_lookup_field(attributes: List[Any]) -> str:
        if not attributes:
            return ""
        non_id_attrs = [attr for attr in attributes if not getattr(attr, "is_id", False)]
        selected = non_id_attrs[0] if non_id_attrs else attributes[0]
        return getattr(selected, "name", "")

    @staticmethod
    def _sorted_by_name(items: Iterable[Any]) -> List[Any]:
        if not items:
            return []
        # Sort by display_order if available (preserves JSON order), otherwise by name
        return sorted(items, key=lambda item: (getattr(item, "display_order", 999999), getattr(item, "name", "").lower()))

    @staticmethod
    def _sorted_menu_items(items: Iterable[MenuItem]) -> List[MenuItem]:
        if not items:
            return []
        return sorted(items, key=lambda item: (item.label or "").lower())

    @staticmethod
    def _sanitize_class_list(class_list) -> Optional[List[str]]:
        if not class_list:
            return None
        names: List[str] = []
        for item in class_list:
            if isinstance(item, dict):
                name = item.get("name")
                if name:
                    names.append(str(name))
            elif isinstance(item, str):
                names.append(item)
        return names or None

    @staticmethod
    def _sanitize_attributes(attributes) -> Dict[str, Any]:
        if not isinstance(attributes, dict):
            return {}
        safe: Dict[str, Any] = {}
        for key, value in attributes.items():
            if isinstance(value, (str, int, float, bool)) or value is None:
                safe[key] = value
            else:
                safe[key] = str(value)
        return safe

    @staticmethod
    def _map_component_type(element: ViewComponent) -> str:
        if isinstance(element, ViewContainer):
            return "container"
        if isinstance(element, Text):
            return "text"
        if isinstance(element, Image):
            return "image"
        if isinstance(element, Link):
            return "link"
        if isinstance(element, Button):
            return "button"
        if isinstance(element, InputField):
            return "input"
        if isinstance(element, Form):
            return "form"
        if isinstance(element, EmbeddedContent):
            return "embedded-content"
        if isinstance(element, Menu):
            return "menu"
        if isinstance(element, DataList):
            return "data-list"
        if isinstance(element, LineChart):
            return "line-chart"
        if isinstance(element, BarChart):
            return "bar-chart"
        if isinstance(element, PieChart):
            return "pie-chart"
        if isinstance(element, RadarChart):
            return "radar-chart"
        if isinstance(element, RadialBarChart):
            return "radial-bar-chart"
        if isinstance(element, Table):
            return "table"
        if isinstance(element, MetricCard):
            return "metric-card"
        if isinstance(element, AgentComponent):
            return "agent-component"
        return "component"

    @staticmethod
    def _derive_display_name(element: ViewComponent) -> str:
        title = getattr(element, "title", None)
        if isinstance(title, str) and title.strip():
            return title.strip()
        if isinstance(element, Text):
            content = getattr(element, "content", "")
            if isinstance(content, str):
                trimmed = content.strip()
                if trimmed:
                    return trimmed
        return GuiSerializationMixin._humanize(element.name)

    @staticmethod
    def _humanize(value: Optional[str]) -> str:
        if not value:
            return ""
        return value.replace("_", " ").strip()

    def _collect_all_table_fields(self) -> List[str]:
        """
        Collect all unique field names from FieldColumns across all Tables in the GUI model.
        Recursively searches through all ViewContainers to find nested Tables.
        These will be used to generate dynamic option display logic in the template.
        
        Returns:
            List of unique field names sorted alphabetically
        """
        field_names = set()
        
        def scan_element_for_tables(element):
            """Recursively scan an element and its children for Tables."""
            if isinstance(element, Table):
                # Extract field names from FieldColumns
                for col in getattr(element, "columns", []) or []:
                    if isinstance(col, FieldColumn):
                        field_name = col.field.name if hasattr(col.field, "name") else str(col.field)
                        field_names.add(field_name)
            
            # Recursively scan children if this is a ViewContainer
            if isinstance(element, ViewContainer):
                for child in element.view_elements:
                    scan_element_for_tables(child)
        
        # Iterate through all modules and screens to find Tables (including nested ones)
        for module in self.gui_model.modules:
            for screen in module.screens:
                for element in screen.view_elements:
                    scan_element_for_tables(element)
        
        return sorted(field_names)

    def _get_used_component_types(self) -> set:
        """
        Scan the GUI model to determine which component types are actually used.
        Recursively searches through all ViewContainers to find nested components.
        
        Returns:
            Set of component type names (e.g., 'LineChart', 'BarChart', 'Table', etc.)
        """
        used_types = set()
        
        def scan_element(element):
            """Recursively scan an element and its children for component types."""
            # Map component instances to their type names
            if isinstance(element, LineChart):
                used_types.add('LineChart')
            elif isinstance(element, BarChart):
                used_types.add('BarChart')
            elif isinstance(element, PieChart):
                used_types.add('PieChart')
            elif isinstance(element, RadarChart):
                used_types.add('RadarChart')
            elif isinstance(element, RadialBarChart):
                used_types.add('RadialBarChart')
            elif isinstance(element, Table):
                used_types.add('Table')
            elif isinstance(element, MetricCard):
                used_types.add('MetricCard')
            
            # Recursively scan children if this is a ViewContainer
            if isinstance(element, ViewContainer):
                for child in element.view_elements:
                    scan_element(child)
        
        # Scan all screens and their children
        for module in self.gui_model.modules:
            for screen in module.screens:
                for element in screen.view_elements:
                    scan_element(element)
        
        return used_types

    @staticmethod
    def _enum_value(value: Any) -> Optional[str]:
        if value is None:
            return None
        if isinstance(value, Enum):
            return value.value if hasattr(value, "value") else value.name
        return str(value)
    
    def _get_entity_name_from_id(self, entity_id: str) -> Optional[str]:
        """Get entity name from the structural model by ID"""
        if not hasattr(self, 'model') or not self.model:
            return None
        
        # Search for class with matching name or component_id
        for cls in self.model.get_classes():
            if cls.name == entity_id or getattr(cls, 'component_id', None) == entity_id:
                return cls.name
        
        # If not found, return the ID as-is (it might already be a name)
        return entity_id
    
    @staticmethod
    def _infer_type(value: Any) -> str:
        """Infer parameter type from value"""
        if isinstance(value, bool):
            return 'boolean'
        elif isinstance(value, int):
            return 'number'
        elif isinstance(value, float):
            return 'number'
        elif isinstance(value, str):
            return 'string'
        else:
            return 'string'

    @staticmethod
    def _convert_style_keys(style: Dict[str, Any]) -> Dict[str, Any]:
        converted: Dict[str, Any] = {}
        for key, value in style.items():
            if value in (None, ""):
                continue
            # Strip !important flags as React inline styles don't support them
            clean_value = str(value).replace(" !important", "").replace("!important", "").strip() if isinstance(value, str) else value
            # Handle gradients - they need to go in backgroundImage, not backgroundColor
            camel_key = GuiSerializationMixin._to_camel_case(key)
            if camel_key == "backgroundImage" and isinstance(clean_value, str):
                # Already correct property for gradients
                converted[camel_key] = clean_value
            elif camel_key == "backgroundColor" and isinstance(clean_value, str) and ("linear-gradient" in clean_value or "radial-gradient" in clean_value):
                # Move gradients from backgroundColor to backgroundImage
                converted["backgroundImage"] = clean_value
            else:
                converted[camel_key] = clean_value
        return converted

    @staticmethod
    def _to_camel_case(name: str) -> str:
        if not name:
            return name
        parts = name.replace("-", "_").split("_")
        first, *rest = parts
        camel = first
        for part in rest:
            if part:
                camel += part[0].upper() + part[1:]
        return camel

    @staticmethod
    def _filter_style(style: Dict[str, Any]) -> Dict[str, Any]:
        filtered: Dict[str, Any] = {}
        for key, value in style.items():
            if value is None:
                continue
            if isinstance(value, str) and value == "":
                continue
            filtered[key] = value
        return filtered

    @staticmethod
    def _clean_dict(data: Dict[str, Any], preserve_keys: set = None) -> Dict[str, Any]:
        """Clean dict by removing empty/None values.
        
        Args:
            data: Dictionary to clean
            preserve_keys: Set of keys to preserve even if their value is empty
        """
        if preserve_keys is None:
            preserve_keys = {"components", "children"}  # Always preserve these for TypeScript compatibility
        
        cleaned: Dict[str, Any] = {}
        for key, value in data.items():
            if isinstance(value, bool):
                cleaned[key] = value
                continue
            # Preserve specified keys even if empty (for TypeScript type safety)
            if key in preserve_keys:
                cleaned[key] = value if value is not None else []
                continue
            if value in (None, "", [], {}, ()):
                continue
            cleaned[key] = value
        return cleaned




import json
from datetime import date, datetime, time, timedelta
from typing import Any, Dict, List

from besser.BUML.metamodel.object import ObjectModel
from besser.BUML.metamodel.object.object import DataValue
from besser.generators import GeneratorInterface


class JSONObjectGenerator(GeneratorInterface):
    """Generate JSON documents from BESSER ObjectModel instances."""

    def __init__(self, model: ObjectModel, output_dir: str = None):
        if not isinstance(model, ObjectModel):
            raise TypeError("JSONObjectGenerator expects an ObjectModel instance")
        super().__init__(model, output_dir)

    def generate(self):
        payload = self._build_document()
        file_name = f"{self._sanitize_name(self.model.name or 'object_model')}.json"
        file_path = self.build_generation_path(file_name)
        with open(file_path, "w", encoding="utf-8") as outfile:
            json.dump(payload, outfile, indent=2, ensure_ascii=False)

    def _build_document(self) -> Dict[str, Any]:
        objects = sorted(self.model.objects, key=lambda obj: obj.name_.lower())
        relationships = self._build_relationship_index(objects)

        items = []
        for obj in objects:
            entry: Dict[str, Any] = {
                "id": obj.name_,
                "class": obj.classifier.name,
            }

            attributes = self._serialize_attributes(obj)
            if attributes:
                entry["attributes"] = attributes

            obj_relationships = relationships.get(obj.name_)
            if obj_relationships:
                entry["relationships"] = obj_relationships

            items.append(entry)

        document: Dict[str, Any] = {
            "name": self.model.name,
            "objects": items,
        }

        description = getattr(getattr(self.model, "metadata", None), "description", None)
        if description:
            document["description"] = description

        return document

    def _serialize_attributes(self, obj) -> Dict[str, Any]:
        data: Dict[str, Any] = {}
        for slot in getattr(obj, "slots", []) or []:
            attribute = getattr(slot, "attribute", None)
            if not attribute:
                continue
            value = getattr(slot, "value", None)
            data[attribute.name] = self._serialize_value(value)
        return data

    def _build_relationship_index(self, objects) -> Dict[str, Dict[str, List[str]]]:
        rel_index: Dict[str, Dict[str, List[str]]] = {obj.name_: {} for obj in objects}
        link_set = getattr(self.model, "links", set()) or []

        for link in link_set:
            association_name = getattr(getattr(link, "association", None), "name", None)
            connections = getattr(link, "connections", []) or []
            for idx, end in enumerate(connections):
                source_obj = getattr(end, "object", None)
                if not source_obj:
                    continue
                target_names = []
                for other_idx, other_end in enumerate(connections):
                    if other_idx == idx:
                        continue
                    target_obj = getattr(other_end, "object", None)
                    if target_obj:
                        target_names.append(target_obj.name_)

                if not target_names:
                    continue

                rel_name = getattr(getattr(end, "association_end", None), "name", None) or association_name
                if not rel_name:
                    rel_name = "association"

                existing = rel_index.setdefault(source_obj.name_, {}).setdefault(rel_name, [])
                for target in target_names:
                    if target not in existing:
                        existing.append(target)

        return {key: value for key, value in rel_index.items() if value}

    def _serialize_value(self, slot_value):
        value = slot_value
        if isinstance(slot_value, DataValue):
            value = slot_value.value

        literal_name = getattr(value, "name", None)
        if literal_name and value.__class__.__name__ == "EnumerationLiteral":
            return literal_name

        if isinstance(value, (datetime, date, time)):
            return value.isoformat()
        if isinstance(value, timedelta):
            return value.total_seconds()
        if isinstance(value, set):
            return [self._serialize_value(item) for item in sorted(value, key=lambda item: str(item))]
        if isinstance(value, list):
            return [self._serialize_value(item) for item in value]
        if isinstance(value, tuple):
            return [self._serialize_value(item) for item in value]
        if hasattr(value, "name_"):
            return value.name_
        return value

    def _sanitize_name(self, name: str) -> str:
        cleaned = (name or "object_model").strip().replace(" ", "_")
        return cleaned or "object_model"



# Generated from D:/Projects/BESSER/besser/BUML/notations/action_language/BESSERActionLanguage.g4 by ANTLR 4.13.2
from antlr4 import *
# from win32comext.mapi.mapi import FORCE_SAVE


from .BESSERActionLanguageLexer import BESSERActionLanguageLexer
from .helpers import functions_for_sequence_type, base_classes, UnknownClassifier
from ...metamodel.action_language.action_language import FunctionDefinition, Parameter, While, DoWhile, For, Iterator, \
    Statement, Block, Condition, AnyType, SequenceType, RealType, StringType, IntType, BoolType, Assignment, Ternary, \
    Or, And, Equal, LessEq, InstanceOf, Plus, Remain, Mult, Not, UnaryMinus, Cast, NullCoalessing, ArrayAccess, \
    FieldAccess, This, New, IntLiteral, StringLiteral, BoolLiteral, RealLiteral, NullLiteral, EnumLiteral, \
    SequenceLiteral, RangeLiteral, NameDecl, Reference, ImplicitDecl, Unequal, Greater, GreaterEq, Less, Minus, Div, \
    ObjectType, EnumType, MethodCall, StandardLibCall, Concatenation, Multiplicity, Return, ExplicitDecl, FunctionType, \
    ProcedureCall, OptionalType, Nothing
from ...metamodel.structural import DomainModel, Class, Enumeration

if "." in __name__:
    from .BESSERActionLanguageParser import BESSERActionLanguageParser
else:
    from BESSERActionLanguageParser import BESSERActionLanguageParser


def parse_bal(domain_model: DomainModel, the_class: Class, code:str):
    if "." in __name__:
        from .BESSERActionLanguageParser import BESSERActionLanguageParser
    else:
        from BESSERActionLanguageParser import BESSERActionLanguageParser
    lexer = BESSERActionLanguageLexer(InputStream(code))
    stream = CommonTokenStream(lexer)
    parser = BESSERActionLanguageParser(stream)

    tree = parser.function_definition()

    visitor = BESSERActionLanguageVisitor(domain_model, the_class)
    return visitor.visit(tree)


class BESSERActionLanguageVisitor(ParseTreeVisitor):

    def __init__(self, domain_model: DomainModel, context_class: Class):
        self.__symbols:dict[str, NameDecl] = {}
        self.__model = domain_model
        self.__method_class = context_class
        self.__current_type = ObjectType(context_class)
        self.__is_root = True
        self.__do_assign = False


    # Visit a parse tree produced by BESSERActionLanguageParser#function_definition.
    def visitFunction_definition(self, ctx:BESSERActionLanguageParser.Function_definitionContext):
        is_root = self.__is_root
        self.__is_root = False
        parameters:list[Parameter] = []
        for param in ctx.params:
            parameters.append(self.visit(param))
        for param in ctx.paramsDefault:
            parameters.append(self.visit(param))

        if is_root and self.__method_class is not None:
            parameters.insert(0, Parameter("self"))

        definition = FunctionDefinition(
            ctx.name.text,
            parameters,
            self.visit(ctx.return_type),
            self.visit(ctx.body)
        )

        self.__current_type = None

        if not is_root:
            self.__symbols[ctx.name.text] = definition

        return definition


    # Visit a parse tree produced by BESSERActionLanguageParser#parameter.
    def visitParameter(self, ctx:BESSERActionLanguageParser.ParameterContext):
        the_type = self.visit(ctx.declared_type)
        param =  Parameter(ctx.name.text, the_type, None)
        self.__symbols[ctx.name.text] = param
        self.__current_type = None
        return param


    # Visit a parse tree produced by BESSERActionLanguageParser#parameterWithDefault.
    def visitParameterWithDefault(self, ctx: BESSERActionLanguageParser.ParameterWithDefaultContext):
        the_type = self.visit(ctx.declared_type)
        default = self.visit(ctx.expr)
        param = Parameter(ctx.name.text, OptionalType(the_type), default)
        self.__symbols[ctx.name.text] = param
        self.__current_type = None
        return param


    # Visit a parse tree produced by BESSERActionLanguageParser#statements.
    def visitStatements(self, ctx:BESSERActionLanguageParser.StatementsContext):
        return self.visit(ctx.getChild(0))


    # Visit a parse tree produced by BESSERActionLanguageParser#cond_loop.
    def visitCond_loop(self, ctx:BESSERActionLanguageParser.Cond_loopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#while.
    def visitWhile(self, ctx:BESSERActionLanguageParser.WhileContext):
        return While(self.visit(ctx.cond), self.visit(ctx.body))


    # Visit a parse tree produced by BESSERActionLanguageParser#do_while.
    def visitDo_while(self, ctx:BESSERActionLanguageParser.Do_whileContext):
        return DoWhile(self.visit(ctx.cond), self.visit(ctx.body))


    # Visit a parse tree produced by BESSERActionLanguageParser#for.
    def visitFor(self, ctx:BESSERActionLanguageParser.ForContext):
        iterators:set[Iterator] = set()
        for it in ctx.iterators:
            iterators.add(self.visit(it))
        return For(iterators, self.visit(ctx.body))


    # Visit a parse tree produced by BESSERActionLanguageParser#iterator.
    def visitIterator(self, ctx:BESSERActionLanguageParser.IteratorContext):
        self.__do_assign = True
        var_name = self.visit(ctx.var_name)
        self.__do_assign = False
        sequence = self.visit(ctx.sequence)
        if isinstance(self.__current_type, SequenceType):
            var_name.declared_type = self.__current_type.elementsType
            var_name.multiplicity = Multiplicity(True, False)
        self.__current_type = None
        return Iterator(var_name, sequence)


    # Visit a parse tree produced by BESSERActionLanguageParser#conditional_branch.
    def visitConditional_branch(self, ctx:BESSERActionLanguageParser.Conditional_branchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#block.
    def visitBlock(self, ctx:BESSERActionLanguageParser.BlockContext):
        children: list[Statement] = list()
        for elem in ctx.stmts:
            children.append(self.visit(elem))
        self.__current_type = None
        return Block(children)


    # Visit a parse tree produced by BESSERActionLanguageParser#condition.
    def visitCondition(self, ctx:BESSERActionLanguageParser.ConditionContext):
        cond = self.visit(ctx.cond)
        then = self.visit(ctx.then)
        elze = None
        if ctx.elze is not None:
            elze = self.visit(ctx.elze)
        self.__current_type = None
        return Condition(cond, then, elze)


    # Visit a parse tree produced by BESSERActionLanguageParser#return.
    def visitReturn(self, ctx: BESSERActionLanguageParser.ReturnContext):
        expr = self.visit(ctx.expr)
        self.__current_type = None
        return Return(expr)


    # Visit a parse tree produced by BESSERActionLanguageParser#type.
    def visitType(self, ctx:BESSERActionLanguageParser.TypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#single_type.
    def visitSingle_type(self, ctx:BESSERActionLanguageParser.Single_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#sequence_type.
    def visitSequence_type(self, ctx: BESSERActionLanguageParser.Sequence_typeContext):
        return SequenceType(self.visit(ctx.the_type))

    # Visit a parse tree produced by BESSERActionLanguageParser#optional_type.
    def visitOptional_type(self, ctx: BESSERActionLanguageParser.Optional_typeContext):
        return OptionalType(self.visit(ctx.the_type))

    # Visit a parse tree produced by BESSERActionLanguageParser#function_type.
    def visitFunction_type(self, ctx: BESSERActionLanguageParser.Function_typeContext):
        params_types = []
        for param in ctx.params_type:
            params_types.append(self.visit(param))
        for param in ctx.params_type_opt:
            params_types.append(self.visit(param))
        return FunctionType(params_types, self.visit(ctx.return_type))


    # Visit a parse tree produced by BESSERActionLanguageParser#any_type.
    def visitAny_type(self, ctx:BESSERActionLanguageParser.Any_typeContext):
        return AnyType()


    # Visit a parse tree produced by BESSERActionLanguageParser#classifier_type.
    def visitClassifier_type(self, ctx:BESSERActionLanguageParser.Classifier_typeContext):
        model_type = self.__model.get_type_by_name(ctx.name.text)
        if isinstance(model_type, Class):
            return ObjectType(model_type)
        elif isinstance(model_type, Enumeration):
            return EnumType(model_type)
        return UnknownClassifier(ctx.name.text)


    # Visit a parse tree produced by BESSERActionLanguageParser#real_type.
    def visitReal_type(self, ctx:BESSERActionLanguageParser.Real_typeContext):
        return RealType()


    # Visit a parse tree produced by BESSERActionLanguageParser#string_type.
    def visitString_type(self, ctx:BESSERActionLanguageParser.String_typeContext):
        return StringType()


    # Visit a parse tree produced by BESSERActionLanguageParser#int_type.
    def visitInt_type(self, ctx:BESSERActionLanguageParser.Int_typeContext):
        return IntType()


    # Visit a parse tree produced by BESSERActionLanguageParser#bool_type.
    def visitBool_type(self, ctx:BESSERActionLanguageParser.Bool_typeContext):
        return BoolType()

    # Visit a parse tree produced by BESSERActionLanguageParser#nothing.
    def visitNothing(self, ctx: BESSERActionLanguageParser.NothingContext):
        return Nothing()


    # Visit a parse tree produced by BESSERActionLanguageParser#expression.
    def visitExpression(self, ctx:BESSERActionLanguageParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#assign_target.
    def visitAssign_target(self, ctx:BESSERActionLanguageParser.Assign_targetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#explicit_declaration.
    def visitExplicit_declaration(self, ctx: BESSERActionLanguageParser.Explicit_declarationContext):
        the_type = self.visit(ctx.declared_type)
        mult = Multiplicity(False, isinstance(the_type, SequenceType))
        decl = ExplicitDecl(ctx.name.text, the_type, mult)
        self.__symbols[ctx.name.text] = decl
        self.__current_type = None
        return decl


    # Visit a parse tree produced by BESSERActionLanguageParser#assignment.
    def visitAssignment(self, ctx:BESSERActionLanguageParser.AssignmentContext):
        if ctx.target is None or ctx.assignee is None:
            return  self.visitChildren(ctx)
        self.__do_assign = True
        target = self.visit(ctx.target)
        self.__do_assign = False
        assignee = self.visit(ctx.assignee)
        if isinstance(target, NameDecl):
            if target.declared_type is None:
                target.declared_type = self.__current_type
                target.multiplicity = Multiplicity(False, isinstance(self.__current_type, SequenceType))
        return Assignment(target, assignee)


    # Visit a parse tree produced by BESSERActionLanguageParser#ternary.
    def visitTernary(self, ctx:BESSERActionLanguageParser.TernaryContext):
        if ctx.expr is None:
            return self.visitChildren(ctx)
        expr = self.visit(ctx.expr)
        elze = self.visit(ctx.elze)
        then = self.visit(ctx.then)
        return Ternary(expr, then, elze)


    # Visit a parse tree produced by BESSERActionLanguageParser#boolean.
    def visitBoolean(self, ctx:BESSERActionLanguageParser.BooleanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#or.
    def visitOr(self, ctx:BESSERActionLanguageParser.OrContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        return Or(self.visit(ctx.left), self.visit(ctx.right))


    # Visit a parse tree produced by BESSERActionLanguageParser#and.
    def visitAnd(self, ctx:BESSERActionLanguageParser.AndContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        return And(self.visit(ctx.left), self.visit(ctx.right))


    # Visit a parse tree produced by BESSERActionLanguageParser#equality.
    def visitEquality(self, ctx:BESSERActionLanguageParser.EqualityContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        self.__current_type = BoolType()
        match ctx.op.text:
            case "==":
                return Equal(left, right)
            case "!=":
                return Unequal(left, right)


    # Visit a parse tree produced by BESSERActionLanguageParser#comparison.
    def visitComparison(self, ctx:BESSERActionLanguageParser.ComparisonContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        self.__current_type = BoolType()
        match ctx.op.text:
            case "<":
                return Less(left, right)
            case "<=":
                return LessEq(left, right)
            case ">":
                return Greater(left, right)
            case ">=":
                return GreaterEq(left, right)


    # Visit a parse tree produced by BESSERActionLanguageParser#instanceof.
    def visitInstanceof(self, ctx:BESSERActionLanguageParser.InstanceofContext):
        if ctx.instance is None or ctx.the_type is None:
            return self.visitChildren(ctx)
        self.__current_type = BoolType()
        return InstanceOf(self.visit(ctx.instance), self.visit(ctx.the_type))


    # Visit a parse tree produced by BESSERActionLanguageParser#arithmetic.
    def visitArithmetic(self, ctx:BESSERActionLanguageParser.ArithmeticContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#plus_minus.
    def visitPlus_minus(self, ctx:BESSERActionLanguageParser.Plus_minusContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        left = self.visit(ctx.left)
        left_type = self.__current_type
        right = self.visit(ctx.right)
        right_type = self.__current_type
        match ctx.op.text:
            case '+':
                if isinstance(left_type, StringType) or isinstance(right_type, StringType):
                    self.__current_type = StringType
                    return Concatenation(left, right)
                elif isinstance(left_type, RealType):
                    self.__current_type = RealType
                return Plus(left, right)

            case '-':
                if isinstance(left_type, StringType) or isinstance(right_type, StringType):
                    return None
                elif isinstance(left_type, RealType):
                    self.__current_type = RealType
                return Minus(left, right)


    # Visit a parse tree produced by BESSERActionLanguageParser#mult_div.
    def visitMult_div(self, ctx:BESSERActionLanguageParser.Mult_divContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        left = self.visit(ctx.left)
        left_type = self.__current_type
        right = self.visit(ctx.right)
        if isinstance(left_type, RealType):
            self.__current_type = RealType
        match ctx.op.text:
            case '*':
                return Mult(left, right)
            case '/':
                return Div(left, right)


    # Visit a parse tree produced by BESSERActionLanguageParser#remain.
    def visitRemain(self, ctx:BESSERActionLanguageParser.RemainContext):
        if ctx.left is None or ctx.right is None:
            return self.visitChildren(ctx)
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        self.__current_type = IntType()
        return Remain(left, right)


    # Visit a parse tree produced by BESSERActionLanguageParser#primary.
    def visitPrimary(self, ctx:BESSERActionLanguageParser.PrimaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#not.
    def visitNot(self, ctx:BESSERActionLanguageParser.NotContext):
        return Not(self.visit(ctx.expr))


    # Visit a parse tree produced by BESSERActionLanguageParser#minus.
    def visitMinus(self, ctx:BESSERActionLanguageParser.MinusContext):
        return UnaryMinus(self.visit(ctx.expr))


    # Visit a parse tree produced by BESSERActionLanguageParser#cast.
    def visitCast(self, ctx:BESSERActionLanguageParser.CastContext):
        expr = self.visit(ctx.expr)
        as_type = self.visit(ctx.as_)
        self.__current_type = as_type
        return Cast(expr, as_type)


    # Visit a parse tree produced by BESSERActionLanguageParser#NullCoalessing.
    def visitNullCoalessing(self, ctx: BESSERActionLanguageParser.NullCoalessingContext):
        elze = self.visit(ctx.elze)
        nullable = self.visit(ctx.nullable)
        return NullCoalessing(nullable, elze)

    # Visit a parse tree produced by BESSERActionLanguageParser#Select.
    def visitSelect(self, ctx: BESSERActionLanguageParser.SelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#ArrayAccess.
    def visitArrayAccess(self, ctx:BESSERActionLanguageParser.ArrayAccessContext):
        index = self.visit(ctx.index)
        receiver = self.visit(ctx.receiver)
        if isinstance(self.__current_type, SequenceType):
            self.__current_type = self.__current_type.elementsType
        else:
            self.__current_type = None
        return ArrayAccess(receiver, index)


    # Visit a parse tree produced by BESSERActionLanguageParser#FunctionCall.
    def visitFunctionCall(self, ctx:BESSERActionLanguageParser.FunctionCallContext):
        args = list()
        for arg in ctx.args:
            args.append(self.visit(arg))

        receiver = self.visit(ctx.receiver)
        if isinstance(self.__current_type, ObjectType) and self.__current_type.clazz is not None:
            clazz =  self.__current_type.clazz
            method = ctx.name.text
            methods = {m for m in clazz.methods if m.name == method}
            if len(methods) > 0:
                method_obj = next(iter(methods))
                return_type = method_obj.type
                the_type = None
                if isinstance(return_type, Class):
                    the_type = ObjectType(return_type)
                elif isinstance(return_type, Enumeration):
                    the_type = EnumType(return_type)
                elif return_type.name in base_classes:
                    the_type = base_classes[return_type.name]
                self.__current_type = the_type
                return MethodCall(receiver, method_obj, args)
            return MethodCall(receiver, None, args)
        elif isinstance(self.__current_type, SequenceType):
            fns = functions_for_sequence_type(self.__current_type)
            fn_type = fns.get(ctx.name.text)
            return StandardLibCall(receiver, ctx.name.text, fn_type, args)
        else:
            return StandardLibCall(receiver, ctx.name.text, FunctionType(None, AnyType()), args)


    # Visit a parse tree produced by BESSERActionLanguageParser#FieldAccess.
    def visitFieldAccess(self, ctx:BESSERActionLanguageParser.FieldAccessContext):
        receiver = self.visit(ctx.receiver)
        prop = None
        if isinstance(self.__current_type, ObjectType) and self.__current_type.clazz is not None:
            clazz = self.__current_type.clazz
            field = ctx.field.text
            for attr in clazz.all_attributes():
                if attr.name == field:
                    prop = attr
            for end in clazz.all_association_ends():
                if end.name == field:
                    prop = end
        if prop is not None:
            prop_type = prop.type
            the_type = None
            if isinstance(prop_type, Class):
                the_type = ObjectType(prop_type)
            elif isinstance(prop_type, Enumeration):
                the_type = EnumType(prop_type)
            elif prop_type.name in base_classes:
                the_type = base_classes[prop_type.name]

            if prop.multiplicity.max > 1:
                the_type = SequenceType(the_type)
            self.__current_type = the_type
        else:
            self.__current_type = None

        return FieldAccess(receiver, prop)


    # Visit a parse tree produced by BESSERActionLanguageParser#Atom.
    def visitAtom(self, ctx:BESSERActionLanguageParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#field_access.
    def visitField_access(self, ctx:BESSERActionLanguageParser.Field_accessContext):
        receiver = self.visit(ctx.receiver)
        prop = None
        if isinstance(self.__current_type, ObjectType) and self.__current_type.clazz is not None:
            clazz = self.__current_type.clazz
            field = ctx.field.text
            for attr in clazz.all_attributes():
                if attr.name == field:
                    prop = attr
            for end in clazz.all_association_ends():
                if end.name == field:
                    prop = end
        if prop is not None:
            prop_type = prop.type
            the_type = None
            if isinstance(prop_type, Class):
                the_type = ObjectType(prop_type)
            elif isinstance(prop_type, Enumeration):
                the_type = EnumType(prop_type)
            elif prop_type.name in base_classes:
                the_type = base_classes[prop_type.name]

            if prop.multiplicity.max > 1:
                the_type = SequenceType(the_type)
            self.__current_type = the_type
        else:
            self.__current_type = None

        return FieldAccess(receiver, prop)


    # Visit a parse tree produced by BESSERActionLanguageParser#array_access.
    def visitArray_access(self, ctx:BESSERActionLanguageParser.Array_accessContext):
        index = self.visit(ctx.index)
        receiver = self.visit(ctx.receiver)
        if isinstance(self.__current_type, SequenceType):
            self.__current_type = self.__current_type.elementsType
        return ArrayAccess(receiver, index)


    # Visit a parse tree produced by BESSERActionLanguageParser#function_call.
    def visitFunction_call(self, ctx:BESSERActionLanguageParser.Function_callContext):
        args = list()
        for arg in ctx.args:
            args.append(self.visit(arg))

        receiver = self.visit(ctx.receiver)
        if isinstance(self.__current_type, ObjectType)  and self.__current_type.clazz is not None:
            clazz = self.__current_type.clazz
            method = ctx.name.text
            methods = {m for m in clazz.methods if m.name == method}
            if len(methods) > 0:
                method_obj = next(iter(methods))
                return_type = method_obj.type
                the_type = None
                if isinstance(return_type, Class):
                    the_type = ObjectType(return_type)
                elif isinstance(return_type, Enumeration):
                    the_type = EnumType(return_type)
                elif return_type.name in base_classes:
                    the_type = base_classes[return_type.name]
                self.__current_type = the_type
                return MethodCall(receiver, method_obj, args)
            return MethodCall(receiver, None, args)
        elif isinstance(self.__current_type, SequenceType):
            fns = functions_for_sequence_type(self.__current_type)
            fn_type = fns.get(ctx.name.text)
            return StandardLibCall(receiver, ctx.name.text, fn_type, args)
        else:
            return StandardLibCall(receiver, ctx.name.text, FunctionType(None, AnyType()), args)


    # Visit a parse tree produced by BESSERActionLanguageParser#atomic.
    def visitAtomic(self, ctx:BESSERActionLanguageParser.AtomicContext):
        if ctx.expr is not None:
            return self.visit(ctx.expr)
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#procedure_call.
    def visitProcedure_call(self, ctx: BESSERActionLanguageParser.Procedure_callContext):
        args = list()
        for arg in ctx.args:
            args.append(self.visit(arg))

        decl = self.__symbols.get(ctx.name.text)
        if isinstance(decl, FunctionDefinition):
            self.__current_type = decl.return_type
            return ProcedureCall(decl, args)
        elif isinstance(decl, NameDecl):
            f_type = decl.declared_type
            if isinstance(f_type, FunctionType):
                self.__current_type = f_type.return_type
                params = list(map(lambda t: Parameter("[unknown parameter]", t), f_type.params_type))
                f = FunctionDefinition(decl.name, params, f_type.return_type)
                return ProcedureCall(f, args)
            else:
                self.__current_type = AnyType()
                return ProcedureCall(None, args)
        self.__current_type = AnyType()
        f = FunctionDefinition(decl.name)
        return ProcedureCall(f, args)


    # Visit a parse tree produced by BESSERActionLanguageParser#this.
    def visitThis(self, ctx:BESSERActionLanguageParser.ThisContext):
        self.__current_type = ObjectType(self.__method_class)
        return This(self.__method_class)


    # Visit a parse tree produced by BESSERActionLanguageParser#new.
    def visitNew(self, ctx:BESSERActionLanguageParser.NewContext):
        args = list()
        for arg in ctx.args:
            args.append(self.visit(arg))
        obj_type = self.visit(ctx.clazz)
        self.__current_type = obj_type
        return New(obj_type, args)


    # Visit a parse tree produced by BESSERActionLanguageParser#literal.
    def visitLiteral(self, ctx:BESSERActionLanguageParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#single_literal.
    def visitSingle_literal(self, ctx:BESSERActionLanguageParser.Single_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by BESSERActionLanguageParser#int_literal.
    def visitInt_literal(self, ctx:BESSERActionLanguageParser.Int_literalContext):
        self.__current_type = IntType()
        return IntLiteral(int(ctx.value.text))


    # Visit a parse tree produced by BESSERActionLanguageParser#string_literal.
    def visitString_literal(self, ctx:BESSERActionLanguageParser.String_literalContext):
        self.__current_type = StringType()
        return StringLiteral(ctx.value.text[1:-1])


    # Visit a parse tree produced by BESSERActionLanguageParser#bool_literal.
    def visitBool_literal(self, ctx:BESSERActionLanguageParser.Bool_literalContext):
        self.__current_type = BoolType()
        return BoolLiteral(ctx.value.text == "true")


    # Visit a parse tree produced by BESSERActionLanguageParser#real_literal.
    def visitReal_literal(self, ctx:BESSERActionLanguageParser.Real_literalContext):
        self.__current_type = RealType()
        return RealLiteral(float(ctx.value.text))


    # Visit a parse tree produced by BESSERActionLanguageParser#null_literal.
    def visitNull_literal(self, ctx:BESSERActionLanguageParser.Null_literalContext):
        self.__current_type = None
        return NullLiteral()


    # Visit a parse tree produced by BESSERActionLanguageParser#enum_literal.
    def visitEnum_literal(self, ctx:BESSERActionLanguageParser.Enum_literalContext):
        enum: EnumType = self.visit(ctx.enum)
        the_literal = None
        if enum.enum is not None:
            for literal in enum.enum.literals:
                if literal.name == ctx.name.text:
                    the_literal = literal.name
        self.__current_type = enum
        return EnumLiteral(enum, the_literal)


    # Visit a parse tree produced by BESSERActionLanguageParser#sequence_literal.
    def visitSequence_literal(self, ctx:BESSERActionLanguageParser.Sequence_literalContext):
        values = list()
        for val in ctx.values:
            values.append(self.visit(val))
        value_type:SequenceType = self.visit(ctx.the_type)
        self.__current_type = value_type
        return SequenceLiteral(value_type.elementsType, values)


    # Visit a parse tree produced by BESSERActionLanguageParser#range_literal.
    def visitRange_literal(self, ctx:BESSERActionLanguageParser.Range_literalContext):
        self.__current_type = SequenceType(IntType())
        return RangeLiteral(self.visit(ctx.first), self.visit(ctx.last))


    # Visit a parse tree produced by BESSERActionLanguageParser#symbol.
    def visitSymbol(self, ctx:BESSERActionLanguageParser.SymbolContext):
        if ctx.name.text in self.__symbols:
            decl = self.__symbols[ctx.name.text]
            self.__current_type = decl.declared_type
            return Reference(decl)

        if self.__do_assign:
            decl = ImplicitDecl(ctx.name.text, None, None)
            self.__symbols[ctx.name.text] = decl
            return decl

        return Reference(None, symbol=ctx.name.text)



del BESSERActionLanguageParser